//***************************************************************************
// Autor: Jasso Petr (jas063)
// Rok: 2005
//
// Program nasloucha na rozhrani, monitoruje prijate UDP packety a zobrazuje
//  jejich zdroj, velikost, pocet a rychlost
//
// Povinny argument programu je cislo portu k naslouchani
// 
//***************************************************************************

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#include <sys/param.h>

#include "lib.h"

#define MAX_SOURCES 1024

typedef struct t_ip {
          struct sockaddr_in in_addr;	// Adresa
	  ulong packets;	// Pocet paketu
	  ulong d_packets;	// Prirustek packetu
	  ulong bytes;		// Pocet bajtu
	  ulong d_bytes;	// Prirustek bajtu
	};

// Socket
int sock;

//***************************************************************************
// Funkce vypise napovedu a ukonci program
void help(char *name) {
  printf(
"\n"
"  UDP packet monitor. Monitors incoming UDP packets\n"
"\n"
"  Usage: [-h] port-number\n"
"\n"
"    -h  this help\n"
"\n", name);

  exit(0);
}

//***************************************************************************
// Funkce hlida vybrane handly (socket a stdin), pokud prisly data na
//  stdin vraci 2, pokud na socket, vraci 1, jinak 0
int wait_handles(int socket) {
  fd_set read_wait_set;
  // Vynuluji vsechny bity na 0
  FD_ZERO(&read_wait_set);
      
  // Cekam na stdin
  FD_SET(STDIN_FILENO, &read_wait_set);
			
  // Cekam na socket
  FD_SET(socket, &read_wait_set);
	    		    
  struct timeval tv;
  tv.tv_sec = 1;
  tv.tv_usec = 0;
			     
  // Cekam na vybrane handly
  if(select(MAX(socket, STDIN_FILENO) + 1, &read_wait_set, 0, 0, &tv) < 0) {
    printf("Cannot wait for selected handles! Exitting...");
    exit(1);
  }
								      
  // Prisly data na socket
  if(FD_ISSET(socket, &read_wait_set))
    return 1;
					  
  // Prisly data na stdin
  if(FD_ISSET(STDIN_FILENO, &read_wait_set))
    return 2;
						  
  return 0;
} 

//***************************************************************************
// Funkce vypise informace o prijatych packetech
void print_info(t_ip *pckts_data, int size) {
  static time_t last_time = time(NULL);
  time_t cur_time;  
  
  ulong bytes = 0, packets = 0;
  
  cur_time = time(NULL);
  if(cur_time > last_time) {
    
    last_time = cur_time;
    tm *cur_time_tm = localtime(&cur_time);

    printf("Time: %d:%d:%d ", cur_time_tm->tm_hour, cur_time_tm->tm_min, cur_time_tm->tm_sec);
    for(int i = 0; i < 66; i++)
      printf("-");
    printf("\n");
    printf("%20s%15s%15s%15s%15s\n", "From", "Packets", "Bytes", "Speed (kbps)", "Packets/s");


    for(int i = 0; i < size; i++) {
      double speed = (pckts_data[i].d_bytes / (double)1024) * 8;
      
      printf("%14s:%5d%15d%15d%15.5f%15d\n", inet_ntoa(pckts_data[i].in_addr.sin_addr), ntohs(pckts_data[i].in_addr.sin_port), 
        pckts_data[i].packets, pckts_data[i].bytes, speed, pckts_data[i].d_packets);
            
      pckts_data[i].d_bytes = 0;
      pckts_data[i].d_packets = 0;
    }


  }
}

//***************************************************************************
// Funkce vynuluje v seznamu, pocty packetu a data za sekundu
void zero_vals(t_ip *pckts_data, int size) {
  
  for(int i = 0; i < size; i++) {
    pckts_data[i].d_bytes = 0;
    pckts_data[i].d_packets = 0;
  }
}

//***************************************************************************
// Funkce vlozi informace o prijatem packetu do dat
void store(t_ip *pckts_data, int *size, sockaddr_in in_addr, int len) {
  bool found = false;
  int i = 0;
  
  for(int i = 0; i < *size; i++) {
    // Hledam existujici polozku v tabulce
    if((pckts_data[i].in_addr.sin_addr.s_addr == in_addr.sin_addr.s_addr) && (pckts_data[i].in_addr.sin_port == in_addr.sin_port)) {
      found = true;
      pckts_data[i].packets++;
      pckts_data[i].d_packets++;
      pckts_data[i].bytes += len + HEADERS_SIZE;
      pckts_data[i].d_bytes += len + HEADERS_SIZE;
      return;
    }
  }
  
  // Novy zdroj
  pckts_data[*size].in_addr = in_addr;
  
  pckts_data[*size].packets = 1; 
  pckts_data[*size].d_packets = 1;
  pckts_data[*size].bytes = len + HEADERS_SIZE;
  pckts_data[*size].d_bytes = len + HEADERS_SIZE;
  (*size)++;
}

void clean(void) {
  printf("Try to close socket ...\n");
  close(sock);
  printf("Socket closed.\n");
}
    
void catch_sig(int sig) {
  exit(1);
}
      

//***************************************************************************
// Hlavni program
int main(int argn, char **arg) {

  if(argn <= 1) help(*arg);

  int port = 0;
  
  // Nacteni parametru (host, port)
  for(int i = 1; i < argn; i++) {

    if(!strcmp(arg[i], "-h"))
      help(*arg);

    if(*arg[i] != '-') {
      if(!port)
        port = atoi(arg[i]);
    }
  }  //for

  if(!port) {
    printf("Invalid or missing port number!\n");
    help(*arg);
  }

  printf("Listening UDP packets on port: %d\n", port);

  // Vytvorim IPv4 socket
  int sock = socket(AF_INET, SOCK_DGRAM, 0);
  if(sock == -1) {
    printf("Unable to create a socket!\n");
    exit(1);
  }

  // Vytvorim IPv4 adresu - (host, port)
  in_addr addr_any = { INADDR_ANY };

  sockaddr_in my_addr;
  my_addr.sin_family = AF_INET;	//IPv4
  my_addr.sin_port = htons(port);	//port
  my_addr.sin_addr = addr_any;

  // Priradim adresu listen socketu
  if(bind(sock, (const sockaddr*)&my_addr, sizeof(my_addr)) < 0) {
    printf("Bind failed!\n");
    close(sock);
    exit(1);
  }

  // Zachytavam <CTRL-C>
  signal(SIGINT, catch_sig);
  // Zachytavam SIG_PIPE
  signal(SIGPIPE, catch_sig);
    
  atexit(clean);

  printf("Enter 'q' or press CTRL+C for exit ...\n");

  t_ip *in_data = (t_ip*)malloc(sizeof(t_ip) * MAX_SOURCES);
  int data_size = 0;
  // Opakuji
  while(1) {
    char buff[2048];
    bool data = false;
    
    sockaddr_in cl_addr;
    uint lsa = sizeof(cl_addr);
    
    int res = wait_handles(sock);
    
    if(res == 1) {
    
      int len = recvfrom(sock, buff, sizeof(buff), 0, (struct sockaddr*)&cl_addr, &lsa);
      if(len < 0) {
        printf("Unable to read data from socket!\n");
        exit(0);
      }
      // Ulozim informace o prijatych datech
      store(in_data, &data_size, cl_addr, len);
      
      data = true;
    } else if(res == 2) {
             int len = read(STDIN_FILENO, buff, sizeof(buff));
        
	     if(len < 0)
               printf("Unable to read data from stdin.\n");
	     if(!strncasecmp(buff, "q", 1)) {
	       printf("Exitting ...\n");
	       exit(0);
	     }
           }
	   
    // Pokud prisly nejake data, zobrazim informace
    if(data) {
      print_info(in_data, data_size);
      data = false;
    }
    else  
      // Vynuluji packety za sekundu
      zero_vals(in_data, data_size);
  }  //while(1)
  
  close(sock);

  return 0;
}
